/*
 * Copyright 1998, Brown University, Providence, RI
 *
 * Permission to use and modify this software and its documentation for
 * any purpose other than its incorporation into a commercial product is
 * hereby granted without fee.  Permission to copy and distribute this
 * software and its documentation only for non-commercial use is also
 * granted without fee, provided, however, that the above copyright notice
 * appear in all copies, that both that copyright notice and this permission
 * notice appear in supporting documentation, that the name of Brown
 * University not be used in advertising or publicity pertaining to
 * distribution of the software without specific, written prior permission,
 * and that the person doing the distribution notify Brown University of
 * such distributions outside of his or her organization. Brown University
 * makes no representations about the suitability of this software for
 * any purpose.  It is provided "as is" without express or implied warranty.
 * Brown University requests notification of any modifications to this
 * software or its documentation.
 *
 * Send the following redistribution information:
 *
 *      Name:
 *      Organization:
 *      Address (postal and/or electronic):
 *
 * To:
 *      Software Librarian
 *      Computer Science Department, Box 1910
 *      Brown University
 *      Providence, RI 02912
 *
 *              or
 *
 *      brusd@cs.brown.edu
 *
 * We will acknowledge all electronic notifications.
*/


/*
//////////////////////////////////////////////////////////////////////////////
//                                                                          
//  NAME: Mitch Cherniack (mfc@cs.brown.edu)
//  FILE: QL.y
//  PROJ: QLite -> KOLA translator (Ox Input File)
//  DATE: 7/98		                                                    
//
//////////////////////////////////////////////////////////////////////////////
*/
/*  Overview:
  
    This Translator translates queries posed in the language, Query Slim,
    into KOLA.  Query Slim is a superset of the query language, 
    Query Lite, described in:

      http://w3.rchland.ibm.com/~rbaxter/gbofQuery/qryLiteIO/qryLiteIO.html

    Unlike Query Lite, Query Slim permits:

      1.  Greater Expressivity in the SELECT clause

          Query Lite permits only a single identifier token
          in the SELECT clause of a query (corresponding to the identifier
          identifying individual elements in the collection 
          named in the query's FROM clause.  Therefore, Query Lite
		  queries are restricted to "selection queries".  Query Slim
          permits ANY Query Expression to be used in the SELECT
          clause.  This includes:

		  * path expressions (thereby permitting "projection queries" and
             not just "selection queries" to be posed)

		  * arithmetic expressions

          * other queries!! (thus permitting the construction of nested
		     collections)

     2.  Greater Expressivity in the FROM clause

          Query Lite only permits a single named collection to appear
		  in the FROM clause of a query, again limiting posed
		  queries to "selection queries" on named collections.  Query
		  Slim permits:

		  * unbounded numbers of collections to be named in the FROM
		    clause (thus allowing for "join queries")

		  * collections appearing in the FROM clause to be derived
		    rather than globally named.  Thus, such collections might
			be collection-valued attributes or method results (from
			the use of a path expression) or the results of other queries.

     3.  Greater Expressivity in the WHERE clause

          Query Lite only permits predicate expressions to be 
          conjunctions or disjunctions of simple predicates 
          (which compare path expression values to constants).
          Query Slim also permits expression of predicates
		  involving other queries, as in:

		  * membership predicates (e.g.: x.m1()...mn() IN SELECT ...)
		  * quantification predicates (e.g.:  FORALL x IN (SELECT ...): P (X)

    4.  Other additional queries including:

	   a.  Duplicate Eliminating Queries (SELECT DISTINCT vs. SELECT)

	   
    As yet, Query Slim does not include support for:

     1.  Constants of types 
	      short, byte, long, double, float, Dinteger, Dshort, Dlong, Dfloat, DDouble, char, DCharacter, DBooleab

     2.  Simple Predicate Operations: >, <, >=, <= and LIKE on Strings

     3.  Simple Predicate Operations: =, != on Booleans
	      (instead of x.foo() = FALSE, one would write: "NOT (x.foo())")
     
   I am as yet unclear as to the difference between "." and ".." as method
   invocations, and therefore treat them as the same thing.  Does the latter
   identify methods corresponding to table attributes?

   For the time being, I am assuming that all posed queries are type correct as I do no type-checking.
   Queries that are not type correct (e.g., 3 + "hello") may produce
   unpredictable results and may even result in segmentation faults.
   This will be corrected at a later time.

   I am also assuming that the input query passed to the translator has
   done the necessary conversion of variable references into their 
   values.  (i.e., select x from coll x where x.m1() = "+ abc +" such
   that abc is a variable would not be accepted.  Instead, it would
   be assumed that + abc + has been replaced by the value of abc at the time
   of query invocation.

   I am assuming that all variables appearing in the query are global if
   not declared as part of the query (e.g., sel and C1 in 

      "select x from C1 x where sel.compare (x)"

   To simplify parsing, I am requiring sort flags (ASC or DESC) to be declared in queries

*/
/* Idea Behind Translation:  Every Query is Translated into a function on 
    the empty environment, NULL.  An example translation of query q (T [q]) is
    shown below
  
     select struct (one: x, two: y)
     from x in Sens, y in Caps
     where x.reps.cap = y.name
  
   = select struct (one: V_1, two: V_0)
     from Sens, Caps
     where V_1.reps.cap = V_0.name
  
   would get initially translated into:
  
     iterate (eq OPLUS <cap o reps o snd o fst, name o snd>, <snd o fst, snd>) o 
      flat o 
       iterate (K (t), iter (K (t), id) o <id, K (Caps)>) o 
        flat o 
         iterate (K (t), iter (K (t), id) o <id, K (Sens)>) o 
          singleton ! *

   while

     select struct (one: x, two: y, three: z)
     from x in Sens, y in Regs, z in Cities
     where x.reps = y.name AND y.cap = z.name
  
   = select struct (one: V_2, two: V_1, three: V_0)
     from Sens, Regs, Caps
     where V_2.reps = V_1.name AND V_1.cap = V_0.name
  
   would get initially translated into:

     iterate ((eq OPLUS <reps o snd o fst o fst, name o snd o fst>) & (eq OPLUS <cap o snd o fst, name o snd>), (<<snd o fst o fst, snd o fst>, snd>)) o 
	  flat o iterate (K (t), iter (K (t), id) o <id, K (Caps)>) o 
	   flat o iterate (K (t), iter (K (t), id) o <id, K (Regs)>) o 
	    flat o iterate (K (t), iter (K (t), id) o <id, K (Sens)>) o 
         singleton ! *
  
*/   

%token COMMENT INTCON STRINGCON FLOATCON CHARCON TRUEV FALSEV NIL
%token DOT NE LE GE IS
%token NOT AND OR LIKE
%token ELT FLATTEN STRUCT SET UNION INTERSECT EXCEPT 
%token EXISTS FORALL INP SELECT DISTINCT FROM WHERE 
%token MOD ABS COUNT SUM MIN MAX AVG
%token ORDER GROUP BY IDENT ERROR ASC DESC

%left OR AND
%left NOT
%left '=' '>' '<' LE GE NE
%left '-' '+' INP
%left '*' '/' MOD UNION INTERSECT EXCEPT
%left DOT

%{
#include "KGlobals.H"
#include "KOFShiftL.H"
#include "KOFShiftR.H"
#include "KOBAnd.H"
#include "KOBEqu.H"
#include "KOBGE.H"
#include "KOBGT.H"
#include "KOBInvoke.H"
#include "KOBLE.H"
#include "KOBLT.H"
#include "KOBName.H"
#include "KOBNot.H"
#include "KOBOr.H"
#include "KOBPrimitive.H"
#include "KOBVar.H"
#include "KOBool.H"
#include "KOF1.H"
#include "KOF2.H"
#include "KOFAbs.H"
#include "KOFAdd.H"
#include "KOFAvg.H"
#include "KOFCElt.H"
#include "KOFChoose.H"
#include "KOFCompose.H"
#include "KOFCon.H"
#include "KOFConst.H"
#include "KOFCount.H"
#include "KOFDif.H"
#include "KOFDiv.H"
#include "KOFFlatten.H"
#include "KOFFold.H"
#include "KOFID.H"
#include "KOFTID.H"
#include "KOFIns.H"
#include "KOFInt.H"
#include "KOFInverse.H"
#include "KOFIter.H"
#include "KOFIterate.H"
#include "KOFJoin.H"
#include "KOFLSJoin.H"
#include "KOFMax.H"
#include "KOFMin.H"
#include "KOFMinus.H"
#include "KOFMod.H"
#include "KOFMult.H"
#include "KOFNJoin.H"
#include "KOFPCon.H"
#include "KOFPFold.H"
#include "KOFPlus.H"
#include "KOFRSJoin.H"
#include "KOFSet.H"
#include "KOFSingleton.H"
#include "KOFSort.H"
#include "KOFSum.H"
#include "KOFTimes.H"
#include "KOFUni.H"
#include "KOFUnnest.H"
#include "KOFUser.H"
#include "KOFVar.H"
#include "KOFunction.H"
#include "KONameUnion.H"
#include "KONonType.H"
#include "KOOAny.H"
#include "KOODif.H"
#include "KOODiv.H"
#include "KOOFloat.H"
#include "KOOId.H"
#include "KOOIf.H"
#include "KOOInt.H"
#include "KOOInter.H"
#include "KOOInvoke.H"
#include "KOOLink.H"
#include "KOOMinus.H"
#include "KOOMult.H"
#include "KOOName.H"
#include "KOOPair.H"
#include "KOOPlus.H"
#include "KOOSet.H"
#include "KOOStr.H"
#include "KOOTimes.H"
#include "KOOUni.H"
#include "KOOVar.H"
#include "KOObjList.H"
#include "KOObject.H"
#include "KOPCEX.H"
#include "KOPCElt.H"
#include "KOPCFA.H"
#include "KOPCombine.H"
#include "KOPConjunct.H"
#include "KOPConst.H"
#include "KOPDisjunct.H"
#include "KOPEqual.H"
#include "KOPExists.H"
#include "KOPForall.H"
#include "KOPGreater.H"
#include "KOPGreaterEqual.H"
#include "KOPInv.H"
#include "KOPIsnotnull.H"
#include "KOPIsnull.H"
#include "KOPLess.H"
#include "KOPLessEqual.H"
#include "KOPLex.H"
#include "KOPNegate.H"
#include "KOPNotEqual.H"
#include "KOPTimes.H"
#include "KOPUser.H"
#include "KOPVar.H"
#include "KOPredicate.H"
#include "KOTWrap.H"

#include "TYTAny.H"
#include "TYTBag.H"
#include "TYTBool.H"
#include "TYTChar.H"
#include "TYTDC.H"
#include "TYTError.H"
#include "TYTFlt.H"
#include "TYTFunc.H"
#include "TYTInt.H"
#include "TYTObj.H"
#include "TYTPair.H"
#include "TYTPred.H"
#include "TYTSet.H"
#include "TYTStr.H"
#include "TYTTup.H"
#include "TYTUnk.H"
#include "TYTUser.H"
#include "TYTVar.H"
#include "TYType.H"

#include "ENEnvironment.H"
#include "ENEnvEntry.H"

#include <iostream.h>
#include <strings.h>

#include "QLAuxCode.H"

// #include "result.H"

extern int lineno3;
extern int QLlex ();

extern char *QLtext;
int QLerror (char *);

////////////////////////////////////////////////////////////////////////////////
TObj *GetType (Environment *e, char *s)
{
  return e -> GetType (s);
}
////////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////
NonType *SRule (Function *f, Predicate *p)
{
  if (f == NULL) {
	return new BInvoke (p, new OAny ());
  } else {
	return new OInvoke (f, new OAny ());
  } 
}
////////////////////////////////////////////////////////////////////////////////

extern NonType *result;

%}

@attributes {int val;} INTCON
@attributes {char *string;} STRINGCON
@attributes {char *string;} FLOATCON
@attributes {char *string;} CHARCON
@attributes {char *string;} IDENT 

@attributes {NonType *res;} S
@attributes {Function *func; Predicate *pred; Environment *ienv;} Q DPrL PExp 
@attributes {Function *func;  Environment *ienv; Environment *oenv;} QEnv
@attributes {Function *func;  Environment *ienv;} NEQList OrderOpt
@attributes {Predicate *pred;  Environment *ienv;} WhereOpt
@attributes {Function *func;} MethList 
@attributes {Predicate *pred;} Direcn
@attributes {int val;} DistOpt
  
%start S

%%

S	    :  Comment Q

               {
                  Environment *nenv;
		  nenv = new Environment();
	       }

               @{       @i @Q.ienv@ = (nenv -> add_entry ("NULL", NULL));
                        @i result = @S.res@ = SRule (@Q.func@, @Q.pred@);
               @}


Comment :  COMMENT Comment 
        |
        ;


Q       :  INTCON

			 @{ @i @Q.func@ = IntFunc (@INTCON.val@);
			    @i @Q.pred@ = NULL;
			 @}
			 
            /* F:  K (OInt (i)) */
	        
        |  STRINGCON

			 @{ @i @Q.func@ = StrFunc (@STRINGCON.string@);
			    @i @Q.pred@ = NULL;
			 @}

            /* F:  K (OStr (s)) */
	        
        |  FLOATCON

			 @{ @i @Q.func@ = FloatFunc (@FLOATCON.string@);
			    @i @Q.pred@ = NULL;
			 @}

            /* F:  K (OStr (s)) */

        |  NIL

			 @{ @i @Q.func@ = NilFunc ();
			    @i @Q.pred@ = NULL;
			 @}

            /* F:  K (OAny ()) */
			 
        |  TRUEV

			 @{ @i @Q.func@ = NULL;
			    @i @Q.pred@ = new PConst (new BPrimitive (T));
			 @}

             /* F:  K (BPrim (T)) */
		    
	    |  FALSEV

			 @{ @i @Q.func@ = NULL;
			    @i @Q.pred@ = new PConst (new BPrimitive (F));
			 @}
		    
             /* F:  K (BPrim (F)) */

	    |  '(' Q ')'

			 @{ @i @Q.1.ienv@  = @Q.0.ienv@;

			    @i @Q.0.func@ = CopyFunc (@Q.1.func@);
			    @i @Q.0.pred@ = CopyPred (@Q.1.pred@);
			 @}

             /* F:  (Q) */

	    |  '-' Q

			 @{ @i @Q.1.ienv@  = @Q.0.ienv@;

			    @i @Q.0.func@ = NegFunc (@Q.1.func@);
			    @i @Q.0.pred@ = NULL;
			 @}

             /* F:  C (MULT, -1) o Q */

        |    ELT '(' Q ')'

             @{ @i @Q.1.ienv@  = @Q.0.ienv@;

                @i @Q.0.func@ = ElementFunc (@Q.1.func@);
                @i @Q.0.pred@ = NULL;
             @}

            /* F:  SINGLETON o Q */

        |    ABS '(' Q ')'

             @{ @i @Q.1.ienv@  = @Q.0.ienv@;

                @i @Q.0.func@ = AbsFunc (@Q.1.func@);
                @i @Q.0.pred@ = NULL;
             @}

             /* F:  ABS o Q */

         |  Q '+' Q

			 @{ @i @Q.1.ienv@  = @Q.0.ienv@;
			    @i @Q.2.ienv@  = @Q.0.ienv@;			 

			    @i @Q.0.func@ = PlusFunc (@Q.1.func@, @Q.2.func@);
			    @i @Q.0.pred@ = NULL;
			 @}

             /* F:  ADD o \<Q, Q\> */

         |  Q '-' Q

		     @{ @i @Q.1.ienv@  = @Q.0.ienv@;
			    @i @Q.2.ienv@  = @Q.0.ienv@;			 
				 
			    @i @Q.0.func@ = MinusFunc (@Q.1.func@, @Q.2.func@);
			    @i @Q.0.pred@ = NULL;
			 @}

             /* F:  SUB o \<Q, Q\> */

         |  Q '*' Q

			 @{ @i @Q.1.ienv@  = @Q.0.ienv@;
			    @i @Q.2.ienv@  = @Q.0.ienv@;			 

			    @i @Q.0.func@ = MultFunc (@Q.1.func@, @Q.2.func@);
			    @i @Q.0.pred@ = NULL;
			 @}

             /* F:  MULT o \<Q, Q\> */

         |  Q '/' Q

			 @{ @i @Q.1.ienv@ = @Q.0.ienv@;
			    @i @Q.2.ienv@ = @Q.0.ienv@;			 

			    @i @Q.0.func@ = DivFunc (@Q.1.func@, @Q.2.func@);
			    @i @Q.0.pred@ = NULL;
			 @}

             /* F:  DIV o \<Q, Q\> */


         |    Q MOD Q

             @{ @i @Q.1.ienv@ = @Q.0.ienv@;
                @i @Q.2.ienv@ = @Q.0.ienv@;                  

                @i @Q.0.func@ = ModFunc (@Q.1.func@, @Q.2.func@);
                @i @Q.0.pred@ = NULL;
             @}

         |  Q '=' Q

			 @{ @i @Q.1.ienv@ = @Q.0.ienv@;
			    @i @Q.2.ienv@ = @Q.0.ienv@;

			    @i @Q.func@  = NULL;
			    @i @Q.pred@  = EqPred (@Q.1.func@, @Q.2.func@);
	         @}

             /* P:  EQ OPLUS \<Q, Q\> */

         |  Q NE Q

			 @{ @i @Q.1.ienv@ = @Q.0.ienv@;
			    @i @Q.2.ienv@ = @Q.0.ienv@;

			    @i @Q.func@  = NULL;
			    @i @Q.pred@  = NeqPred (@Q.1.func@, @Q.2.func@);
		     @}

             /* P:  NEQ OPLUS \<Q, Q\> */

         |  Q '<' Q

			 @{ @i @Q.1.ienv@ = @Q.0.ienv@;
			    @i @Q.2.ienv@ = @Q.0.ienv@;

			    @i @Q.func@  = NULL;
			    @i @Q.pred@  = LTPred (@Q.1.func@, @Q.2.func@);
	         @}
		    
             /* P:  LT OPLUS \<Q, Q\> */

         |  Q LE Q

			 @{ @i @Q.1.ienv@ = @Q.0.ienv@;
			    @i @Q.2.ienv@ = @Q.0.ienv@;

			    @i @Q.func@  = NULL;
			    @i @Q.pred@  = LEQPred (@Q.1.func@, @Q.2.func@);
		     @}
		    
             /* P:  LEQ OPLUS \<Q, Q\> */

         |  Q '>' Q

			 @{ @i @Q.1.ienv@ = @Q.0.ienv@;
			    @i @Q.2.ienv@ = @Q.0.ienv@;

			    @i @Q.func@  = NULL;
			    @i @Q.pred@  = GTPred (@Q.1.func@, @Q.2.func@);
		     @}

             /* P:  GT OPLUS \<Q, Q\> */
		    
         |  Q GE Q

			 @{ @i @Q.1.ienv@ = @Q.0.ienv@;
			    @i @Q.2.ienv@ = @Q.0.ienv@;

			    @i @Q.func@  = NULL;
			    @i @Q.pred@  = GEQPred (@Q.1.func@, @Q.2.func@);
	         @}

             /* P:  GEQ OPLUS \<Q, Q\> */

         |  NOT Q

	    	 @{ @i @Q.1.ienv@  = @Q.0.ienv@;

			    @i @Q.0.func@ = NULL;
			    @i @Q.0.pred@ = NotPred (@Q.1.pred@);
			 @}			 
		    
             /* P:  ~ (Q) */

         |  Q AND Q

			 @{ @i @Q.1.ienv@  = @Q.0.ienv@;
			    @i @Q.2.ienv@  = @Q.0.ienv@;

			    @i @Q.0.func@ = NULL;
			    @i @Q.0.pred@ = AndPred (@Q.1.pred@, @Q.2.pred@);
			 @}

             /* P:  Q & Q */
		    
         |  Q OR Q

			 @{ @i @Q.1.ienv@  = @Q.0.ienv@;
			    @i @Q.2.ienv@  = @Q.0.ienv@;

			    @i @Q.0.func@ = NULL;
			    @i @Q.0.pred@ = OrPred (@Q.1.pred@, @Q.2.pred@);
			 @}

             /* P:  Q | Q */

         |  FLATTEN '(' Q ')'

             @{ @i @Q.1.ienv@  = @Q.0.ienv@;

                @i @Q.0.func@ = FlattenFunc (@Q.1.func@);
                @i @Q.0.pred@ = NULL;
             @}

             /* F:  FLAT o Q */

         |  Q UNION Q

             @{ @i @Q.1.ienv@  = @Q.0.ienv@;
                @i @Q.2.ienv@  = @Q.0.ienv@;                         

                @i @Q.0.func@ = UniFunc (@Q.1.func@, @Q.2.func@);
                @i @Q.0.pred@ = NULL;
             @}

             /* F:  UNI (=) o \<Q, Q\> */

         |  Q INTERSECT Q

             @{ @i @Q.1.ienv@  = @Q.0.ienv@;
                @i @Q.2.ienv@  = @Q.0.ienv@;                         

                @i @Q.0.func@ = InterFunc (@Q.1.func@, @Q.2.func@);
                @i @Q.0.pred@ = NULL;
             @}
                    
            /* F:  INT (=) o \<Q, Q\> */

         |  Q EXCEPT Q

             @{ @i @Q.1.ienv@  = @Q.0.ienv@;
                @i @Q.2.ienv@  = @Q.0.ienv@;                         

                @i @Q.0.func@ = DifFunc (@Q.1.func@, @Q.2.func@);
                @i @Q.0.pred@ = NULL;
             @}

             /* F:  DIF (=) o \<Q, Q\> */

         |  Q INP Q

			 @{ @i @Q.1.ienv@ = @Q.0.ienv@;
			    @i @Q.2.ienv@ = @Q.0.ienv@;

			    @i @Q.func@ = NULL;
			    @i @Q.pred@ = InPred (@Q.1.func@, @Q.2.func@);
		     @}

             /* P:  EX (=) OPLUS \<Q, Q\> */

         |  COUNT '(' Q ')'

             @{ @i @Q.1.ienv@  = @Q.0.ienv@;

                @i @Q.0.func@ = COUNTFunc (@Q.1.func@);
                @i @Q.0.pred@ = NULL;
             @}

             /* F: FOLD (0, K (1), add) o Q */

         |  SUM '(' Q ')'

             @{ @i @Q.1.ienv@  = @Q.0.ienv@;

                @i @Q.0.func@ = SUMFunc (@Q.1.func@);
                @i @Q.0.pred@ = NULL;
             @}

             /* F: FOLD (0, id, add) o Q */

         |  MIN '(' Q ')'

             @{ @i @Q.1.ienv@  = @Q.0.ienv@;

                @i @Q.0.func@ = MINFunc (@Q.1.func@);
                @i @Q.0.pred@ = NULL;
             @}

             /* F: PFOLD (id, con (lt, fst, snd)) o Q */

         |  MAX '(' Q ')'

             @{ @i @Q.1.ienv@  = @Q.0.ienv@;

                @i @Q.0.func@ = MAXFunc (@Q.1.func@);
                @i @Q.0.pred@ = NULL;
             @}

             /* F: PFOLD (id, con (lt, snd, fst)) o Q */

         |  AVG '(' Q ')'

             @{ @i @Q.1.ienv@  = @Q.0.ienv@;

                @i @Q.0.func@ = AVGFunc (@Q.1.func@);
                @i @Q.0.pred@ = NULL;
             @}

             /* F: (pcon (C (neq, 0) OPLUS snd, div) o  
                             FOLD ([0, 0], \<id, K (1)\>, (add x add))) o Q */ 


         |  Q IS NIL

			 @{ @i @Q.1.ienv@ = @Q.0.ienv@;

			    @i @Q.func@ = NULL;
			    @i @Q.pred@ = IsNullPred (@Q.1.func@);
		     @}

             /* P:  ISNULL OPLUS Q */

         |  Q IS NOT NIL

			 @{ @i @Q.1.ienv@ = @Q.0.ienv@;

			    @i @Q.func@ = NULL;
			    @i @Q.pred@ = IsNotNullPred (@Q.1.func@);
		     @}

             /* P:  ISNOTNULL OPLUS Q */

         |  FORALL IDENT INP Q ':' '(' Q ')'

		     @{ @i @Q.1.ienv@  = @Q.0.ienv@;
			    @i @Q.2.ienv@  = QuanEnv (@Q.0.ienv@, @IDENT.string@);

			    @i @Q.0.func@ = NULL;
			    @i @Q.0.pred@ = FORALLPred (@Q.1.func@, @Q.2.pred@);
			 @}
		    
             /* P: FA (Q0) OPLUS <id, Q1> */

         |  EXISTS IDENT INP Q ':' '(' Q ')'

		     @{ @i @Q.1.ienv@  = @Q.0.ienv@;
			    @i @Q.2.ienv@  = QuanEnv (@Q.0.ienv@, @IDENT.string@);

			    @i @Q.0.func@ = NULL;
			    @i @Q.0.pred@ = EXISTSPred (@Q.1.func@, @Q.2.pred@);
			 @}
		    
             /* P: EX (Q0) OPLUS <id, Q1> */

         |  SELECT DistOpt Q FROM QEnv WhereOpt OrderOpt

			 @{ @i @QEnv.ienv@        = @Q.0.ienv@;
                @i @Q.1.ienv@         = @QEnv.oenv@;
                @i @WhereOpt.ienv@    = @QEnv.oenv@;
                @i @OrderOpt.ienv@    = @QEnv.oenv@;

			    @i @Q.0.func@         = SFWFunc (@OrderOpt.func@, @Q.1.func@, @QEnv.func@, @WhereOpt.pred@, @DistOpt.val@);
			    @i @Q.0.pred@         = NULL;
			 @}

            /* F: [sort (..) o ] [set o ] iterate (T (Q_1), T (Q_0)) o QEnv o singleton */

         |  STRUCT '(' DPrL ')'
 
	         @{ @i @DPrL.ienv@ = @Q.ienv@;

			    @i @Q.0.func@ = CopyFunc (@DPrL.func@);
			    @i @Q.0.pred@ = NULL;
			 @}
		    
             /* F: DPrL */

         |  IDENT 

			 @{ 
			    @i @Q.func@ = IdentFunc (@IDENT.string@, @Q.ienv@);
			    @i @Q.pred@ = IdentPred (@IDENT.string@, @Q.ienv@);
			 @}


         |  PExp

			 @{ @i @PExp.ienv@   = @Q.0.ienv@;

			    @i @Q.0.func@    = CopyFunc (@PExp.func@);
			    @i @Q.0.pred@    = CopyPred (@PExp.pred@);
			 @}
		    
         ;


PExp    : Q DOT IDENT '(' NEQList ')'

			 @{ @i @Q.ienv@       = @PExp.ienv@;
			    @i @NEQList.ienv@ = @PExp.ienv@;

			    @i @PExp.func@    = Path1Func (@IDENT.string@, @Q.func@, @NEQList.func@);
			    @i @PExp.pred@    = Path1Pred (@IDENT.string@, @Q.func@, @NEQList.func@);
			 @}

           /* F:  IDENT o <Q, QLIST> */

        | Q DOT IDENT 

			 @{ @i @Q.ienv@       = @PExp.ienv@;

			    @i @PExp.func@    = Path2Func (@IDENT.string@, @Q.func@);
			    @i @PExp.pred@    = Path2Pred (@IDENT.string@, @Q.func@);
			 @}

           /* F:  IDENT o Q */

        | Q DOT IDENT '(' ')'

			 @{ @i @Q.ienv@       = @PExp.ienv@;

			    @i @PExp.func@    = Path2Func (@IDENT.string@, @Q.func@);
			    @i @PExp.pred@    = Path2Pred (@IDENT.string@, @Q.func@);
			 @}

           /* F:  IDENT o Q */
        ;

NEQList : NEQList ',' Q

			 @{ @i @NEQList.1.ienv@ = @NEQList.0.ienv@;
			    @i @Q.ienv@         = @NEQList.0.ienv@;

			    @i @NEQList.0.func@ = PairFunc (@NEQList.1.func@, @Q.func@);
			 @}		    

        | Q

			 @{ @i @Q.ienv@         = @NEQList.ienv@;
			    @i @NEQList.func@   = CopyFunc (@Q.func@);
			 @}
		            
        ;

WhereOpt : WHERE Q

             @{ @i @Q.ienv@ = @WhereOpt.ienv@; 
                @i @WhereOpt.pred@ = CopyPred (@Q.pred@);
             @}
   
        |

             @{ @i @WhereOpt.pred@ = new PConst (new BPrimitive (1)); @}
   
        ;


OrderOpt : ORDER BY MethList Direcn

             @{ @i @OrderOpt.func@   = Sort1Func (@MethList.func@, @Direcn.pred@); @}
   
          /* F: SORT (LESS OPLUS (<...<m1, m2>, ..., mn> x <...<m1, m2>, ..., mn>)) */

        |  ORDER BY IDENT Direcn

             @{ @i @OrderOpt.func@ = Sort2Func (@IDENT.string@, @Direcn.pred@); @}
   
          /* F: SORT (LESS OPLUS <K (0), IDENT>) */

        |

             @{ @i @OrderOpt.func@ = new FID (); @}

          /* F: ID */   

        ;



MethList : MethList ',' IDENT '(' ')'

             @{ @i @MethList.0.func@ = PairFunc (@MethList.1.func@, new FUser (@IDENT.string@)); @}

        |  IDENT '(' ')'

             @{ @i @MethList.func@   = new FUser (@IDENT.string@); @}

        ;



Direcn  : ASC

             @{ @i @Direcn.pred@ = new PLess (); @}     

        | DESC

             @{ @i @Direcn.pred@ = new PGreater (); @}     

        ;



DistOpt :  DISTINCT
                         
             @{ @i @DistOpt.val@ = T; @}

        |

             @{ @i @DistOpt.val@ = F; @}

        ;


QEnv	:  Q IDENT

			 @{ @i @Q.ienv@        = @QEnv.ienv@;

                @i @QEnv.func@     = QEnvFunc (@Q.func@);
                @i @QEnv.oenv@     = QEnvEnv (@QEnv.ienv@, @IDENT.string@);
			 @}

            /* F: iterate (K (true), flat o iter (K (true), id) o <id, T (Q)>) */  
                   
        |  QEnv ',' Q IDENT 

			 @{ @i @QEnv.1.ienv@  = @QEnv.0.ienv@;
			    @i @Q.ienv@       = @QEnv.1.oenv@;
			    
                @i @QEnv.0.func@  = QEnvFunc2 (@QEnv.1.func@, @Q.func@);
                @i @QEnv.0.oenv@  = QEnvEnv (@QEnv.1.oenv@, @IDENT.string@);
			 @}

            /* F: iterate (K (true), flat o iter (K (true), id) o <id, T (Q_n)>) o QEnv */ 

        |  IDENT INP Q

			 @{ @i @Q.ienv@        = @QEnv.ienv@;

                @i @QEnv.func@     = QEnvFunc (@Q.func@);
                @i @QEnv.oenv@     = QEnvEnv (@QEnv.ienv@, @IDENT.string@);
			 @}

            /* F: iterate (K (true), flat o iter (K (true), id) o <id, T (Q)>) */  
                   
        |  QEnv ',' IDENT INP Q

			 @{ @i @QEnv.1.ienv@  = @QEnv.0.ienv@;
			    @i @Q.ienv@       = @QEnv.1.oenv@;
			    
                @i @QEnv.0.func@  = QEnvFunc2 (@QEnv.1.func@, @Q.func@);
                @i @QEnv.0.oenv@  = QEnvEnv (@QEnv.1.oenv@, @IDENT.string@);
			 @}

            /* F: iterate (K (true), flat o iter (K (true), id) o <id, T (Q_n)>) o QEnv */ 

        ;


DPrL	:  IDENT ':' Q

			 @{ @i @Q.0.ienv@    = @DPrL.ienv@;

			    @i @DPrL.func@   = CopyFunc (@Q.0.func@);
			    @i @DPrL.pred@   = NULL;

			 @}

        |  DPrL ',' IDENT ':' Q

			 @{ @i @DPrL.1.ienv@  = @DPrL.0.ienv@;
			    @i @Q.ienv@       = @DPrL.0.ienv@;

			    @i @DPrL.0.func@  = PairFunc (@DPrL.1.func@, @Q.func@);
			    @i @DPrL.0.pred@  = NULL;
			 @}

	    ;


%%

int QLerror (char *s)
{
  cerr <<  s << " found at line: " << lineno3 << " (" << QLtext << ")" << "\n\n";
  return 0;
}

/*
	        |    FLOATCON

			 @{ @i @Q.func@ = FloatFunc (@FLOATCON.val@);
			    @i @Q.pred@ = NULL;
			 @}


*/
